from enum import auto
from hashlib import new
from typing import Union, TYPE_CHECKING
from binascii import hexlify

if TYPE_CHECKING:
    from kasten.generator import KastenPacked
from kasten import Kasten
from . import generators

class Block:
    def __init__(
            self, block_hash: str,
            raw_block_data: 'KastenPacked', auto_verify=True):
        generator = generators.AnonVDFGenerator
        bl = Kasten(
            block_hash,
            raw_block_data,
            generator, auto_check_generator=auto_verify)
        self.timestamp = bl.get_timestamp()
        self.metadata = bl.get_metadata()
        self.ttl = self.metadata['ttl']
        self.id = block_hash
        self.type = bl.get_data_type()
        self.data = bl.data
        self.raw = bl.get_packed()


def kasten_to_block(kasten: Kasten):
    return Block(hexlify(kasten.id), kasten.get_packed(), auto_verify=False)




"""
Copyright (C) <2020>  Kevin Froman

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from time import time

from kasten.generator.pack import pack

from ..generators import anonvdf

from ..block import Block
from ..block import kasten_to_block


def create_anonvdf_block(
        block_data: bytes,
        block_type: bytes,
        ttl: int,
        **block_metadata) -> Block:
    try:
        block_data = block_data.encode('utf-8')
    except AttributeError:
        pass
    try:
        block_type = block_type.encode('utf-8')
    except AttributeError:
        pass
    ts = int(time())
    packed = pack(
        block_data, block_type,
        app_metadata=block_metadata, timestamp=ts)
    block_metadata['ttl'] = ttl
    rounds_needed = anonvdf.AnonVDFGenerator.get_rounds_for_ttl_seconds(
        ttl, len(packed))
    packed = pack(
        block_data,
        block_type, app_metadata=block_metadata, timestamp=ts)
    k = anonvdf.AnonVDFGenerator.generate(
        packed, block_metadata['ttl'])
    return kasten_to_block(k)


from setuptools import setup, find_packages

setup(name='onionrblocks',
      version='6.0.0',
      description='Onionr message format',
      author='Kevin Froman',
      author_email='beardog@mailbox.org',
      url='https://onionr.net/',
      packages=find_packages(exclude=['contrib', 'docs', 'tests']),
      install_requires=['kasten>=3.0.0', 'pynacl>=1.4.0'],
      classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
      ],
     )



import time
from binascii import a2b_hex

from kasten import Kasten, generator
from kasten.types import KastenPacked, KastenChecksum
from kasten.exceptions import InvalidID

from mimcvdf import vdf_create, vdf_verify

from onionrblocks.universalrules import check_block_sanity


class NotEnoughRounds(Exception):
    pass

class NoTTLSet(Exception):
    pass


class AnonVDFGenerator(generator.KastenBaseGenerator):

    # 90000 rounds = 1 second (2.8ghz C) = 1 hour storage of 1 MB storage
    # jan 2021 disk price metric = 0.000017 USD per MB
    # mobile MB transfer US/Canada = 0.0125 USD
    # Max cost of block storage + American mobile data transfer = 0.0751 USD
    # Goal: approximately 2 seconds of computation for 1 hour of 1 MB storage

    byte_cost = 10
    second_cost = 4

    @classmethod
    def get_rounds_for_ttl_seconds(cls, seconds: int, size_bytes: int):
        return (seconds * cls.second_cost) + (size_bytes * cls.byte_cost)

    @classmethod
    def get_ttl_seconds_per_rounds(cls, rounds: int, size_bytes: int):
        return (rounds - (size_bytes * cls.byte_cost)) // cls.second_cost

    @classmethod
    def generate(
            cls, packed_bytes: KastenPacked, ttl: int = None) -> Kasten:

        k: Kasten = Kasten(
                    None,
                    packed_bytes, None,
                    auto_check_generator=False)
        try:
            ttl: int = k.get_metadata()['ttl']
        except (KeyError, TypeError) as e:
            raise NoTTLSet(e)

        rounds = AnonVDFGenerator.get_rounds_for_ttl_seconds(ttl, len(packed_bytes))
        check_block_sanity(packed_bytes)
        vdf = vdf_create(
                packed_bytes,
                rounds, dec=True
                ).to_bytes(64, "big")
        return Kasten(vdf, packed_bytes, cls, auto_check_generator=False)

    @staticmethod
    def validate_id(
            hash: KastenChecksum,
            packed_bytes: KastenPacked) -> None:

        check_block_sanity(packed_bytes)
        k: Kasten = Kasten(
                    None,
                    packed_bytes, None,
                    auto_check_generator=False)
        try:
            ttl: int = k.get_metadata()['ttl']
        except (KeyError, TypeError) as e:
            raise NoTTLSet(e)

        ttl = int(ttl)
        rounds = AnonVDFGenerator.get_rounds_for_ttl_seconds(ttl, len(packed_bytes))


        #hash = hash.lstrip(b"0")
        hash = int(hash, 16)

        try:
            hash = int.from_bytes(hash, byteorder="big")
        except TypeError:
            pass
        if not vdf_verify(packed_bytes, hash, rounds):
            raise InvalidID

        test_obj = Kasten(
            None, packed_bytes, None, auto_check_generator=False)
        allowed_age_seconds = AnonVDFGenerator.get_ttl_seconds_per_rounds(
            rounds, len(packed_bytes))
        if time.time() > test_obj.get_timestamp() + allowed_age_seconds:
            raise ValueError(
                f"Block rounds only valid through {allowed_age_seconds}")

        return None